"""

Generate landscapes of interconnected patches for simulating within a spatially explicit neutral model.
Detailed :ref:`here <generate_landscapes>`.

Dispersal probabilities are defined between different patches, and each patch will be contain n individuals.
"""
from __future__ import division

import csv
import math
import os
import sys

import numpy as np

from .map import Map
from .system_operations import check_parent


class Patch(object):
	"""
	Contains a single patch, to which the probability of dispersal to every other patch can be added.
	"""

	def __init__(self, id, density):
		"""
		Generates the patch object with a fixed density and id.

		:param id: the name for the fragment as a unique identifier.
		:param density: the number of individuals that exist (non-spatially) within this patch.
		"""
		self.density = density
		self.id = id
		self.dispersal_probabilities = {}
		self.index = 0

	def __repr__(self):
		"""
		Representation of the object.
		"""
		return "Patch({}, {})".format(self.id, self.density)

	def __str__(self):
		"""
		Prints the patch data to a string.
		:return: string containing the object data
		"""
		str = "Patch class object with id: {}, density: {}," \
			  " index: {} and dispersal probabilities: {}".format(self.id, self.density, self.index,
																  self.dispersal_probabilities)
		return str

	def add_patch(self, patch, probability):
		"""
		Adds dispersal from this patch to another patch object with a set probability. The patch should not already
		have been added.

		.. note:: The probabilities can be relative, as they can be re-scaled to sum to 1 using
				  :func:`re_scale_probabilities`.

		:raises KeyError: if the patch already exists in the dispersal probabilities.

		:raises ValueError: if the dispersal probability is less than 0.

		:param patch: the patch id to disperse to
		:param probability: the probability of dispersal
		"""
		if probability < 0:
			raise ValueError("Dispersal probability should be > 0")
		if patch in self.dispersal_probabilities.keys():
			raise KeyError("Patch {} already exists in dispersal probabilities.".format(patch))
		self.dispersal_probabilities[patch] = probability

	def re_scale_probabilities(self):
		"""
		Re-scales the probabilities so that they sum to 1. Also checks to make sure dispersal from within this patch
		is defined.

		:raises ValueError: if the self dispersal probability has not been defined, or the dispersal probabilities do
							not sum to > 0.
		"""
		has_self = False
		total = 0
		for key, value in self.dispersal_probabilities.items():
			if key == self.id:
				has_self = True
			if value < 0:
				raise ValueError("Dispersal probabilities must be >= 0, currently {}.".format(value))
			total += value
		if total <= 0:
			raise ValueError("Dispersal probabilities must sum to > 0, currently {}.".format(total))
		if not has_self:
			raise ValueError("Self-dispersal probability has not been defined, please add first.")
		for key, value in self.dispersal_probabilities.items():
			self.dispersal_probabilities[key] = value / total


class PatchedLandscape:
	"""
	Landscape made up of a list of patches with dispersal probabilities to each other.
	"""

	def __init__(self, output_fine_map, output_dispersal_map):
		"""
		Creates the patched landscape object, providing the paths to the fine map (for outputting the density) and
		dispersal map.

		:raises IOError: if either output_fine_map or output_dispersal_map exist.

		:param output_fine_map: path to the fine map to create, which will contain the densities
		:param output_dispersal_map: path to the dispersal map, which will contain dispersal probabilities
		"""
		if os.path.exists(output_fine_map):
			raise IOError("Output fine map already exists at {}.".format(output_fine_map))
		if os.path.exists(output_dispersal_map):
			raise IOError("Output dispersal map already exists at {}.".format(output_dispersal_map))
		check_parent(output_fine_map)
		check_parent(output_dispersal_map)
		self.fine_map = Map()
		self.fine_map.file_name = output_fine_map
		self.dispersal_map = Map()
		self.dispersal_map.file_name = output_dispersal_map
		self.patches = {}

	def __repr__(self):
		"""
		Representation of the object.

		:return: string containing the representation of the object.
		"""
		return "PatchedLandscape({}, {})".format(self.fine_map.file_name, self.dispersal_map.file_name)

	def add_patch(self, id, density, self_dispersal=None, dispersal_probabilities=None):
		"""
		Add a patch with the given parameters.

		:param id: the unique reference for the patch
		:param density: the number of individuals that exist in the patch
		:param self_dispersal: the relative probability of dispersal from within the same patch
		:param dispersal_probabilities: dictionary containing all other patches and their relative dispersal
		       probabilities
		"""
		if self.has_patch(id):
			raise KeyError("Patch with id of {} already in patch list.".format(id))
		if density <= 0:
			raise ValueError("Density cannot be less than 0.")
		if self_dispersal is None:
			if dispersal_probabilities is None:
				raise TypeError("Must provide self-dispersal value.")
		patch = Patch(id, density)
		patch.add_patch(id, self_dispersal if self_dispersal is not None else 0.0)
		if dispersal_probabilities:
			if not isinstance(dispersal_probabilities, dict):
				raise TypeError("Dispersal probabilities must be provided as a dictionary of "
								"ids->relative probabilities.")
			if self_dispersal is None:
				if id not in dispersal_probabilities.keys():
					raise KeyError("Must provide self dispersal value either separately or within dictionary.")
				patch.dispersal_probabilities[id] = dispersal_probabilities[id]
			for key, value in dispersal_probabilities.items():
				if key != id:
					patch.add_patch(key, value)
		self.patches[id] = patch

	def has_patch(self, id):
		"""
		Checks if the patches object already contains a patch with the provided id.

		:param id: id to check for in patches

		:return: true if the patch already exists
		"""
		return id in self.patches.keys()

	def add_dispersal(self, source_patch, target_patch, dispersal_probability):
		"""
		Adds a dispersal probability from the source patch to the target patch.

		.. note:: Both the source and target patch should already have been added using :func:`add_patch`.

		:param source_patch: the id of the source patch
		:param target_patch: the id of the target patch
		:param dispersal_probability: the probability of dispersal from source to target
		"""
		if source_patch not in self.patches.keys():
			raise ValueError("Source patch {} has not been added.".format(source_patch))
		if target_patch not in self.patches.keys():
			raise ValueError("Target patch {} has not been added.".format(target_patch))
		self.patches[source_patch].add_patch(target_patch, dispersal_probability)

	def generate_files(self):
		"""
		Re-scales the dispersal probabilities and generates the patches landscape files. These include the fine map file
		containing the densities and the dispersal probability map.

		The fine map file will be dimensions 1xN where N is the number of patches in the landscape.
	 	The dispersal probability map will be dimensions NxN, where dispersal occurs from the y index cell to the x
	 	index cell.
		"""
		map_size = len(self.patches)
		self.fine_map.data = np.zeros(shape=(1, map_size))
		self.dispersal_map.data = np.zeros(shape=(map_size, map_size))
		# Assign indices to each patch
		index = 0
		for key, value in self.patches.items():
			self.patches[key].index = index
			self.patches[key].re_scale_probabilities()
			index += 1
		for k1, src_patch in self.patches.items():
			src_index = src_patch.index
			self.fine_map.data[0, src_index] = src_patch.density
			if len(src_patch.dispersal_probabilities) == 0:  # pragma: no cover
				raise ValueError("No dispersal probabilities supplied in patch {}".format(src_patch.id))
			for k2, dst_patch in self.patches.items():
				dst_index = dst_patch.index
				if k2 not in src_patch.dispersal_probabilities.keys():
					self.dispersal_map.data[src_index, dst_index] = 0.0
				else:
					self.dispersal_map.data[src_index, dst_index] = src_patch.dispersal_probabilities[k2]
		self.fine_map.create(self.fine_map.file_name, datatype=5)
		self.dispersal_map.create(self.dispersal_map.file_name, datatype=6)

	def generate_fragment_csv(self, fragment_csv):
		"""
		Generates a fragment csv for usage within a coalescence simulation, with each patch becomming one fragment on
		the landscape.

		:param fragment_csv: the path to the output csv to create

		:raises IOError: if the output fragment csv already exists
		"""
		if os.path.exists(fragment_csv):
			raise IOError("Output file already exists at {}.".format(fragment_csv))
		check_parent(fragment_csv)
		if sys.version_info[0] < 3:  # pragma: no cover
			infile = open(fragment_csv, "wb")
		else:
			infile = open(fragment_csv, "w", newline='')
		with infile as csv_file:
			csv_writer = csv.writer(csv_file)
			for k1, patch in self.patches.items():
				csv_writer.writerow([k1, patch.index, 0, patch.index, 0, patch.density])

	def generate_from_matrix(self, density_matrix, dispersal_matrix):
		"""
		Generates the patched landscape from the input matrix and writes out to the files.

		.. note:: Uses a slightly inefficient method of generating the full patched landscape, and then writing back out
				  to the map files so that full error-checking is included. A more efficient implementation is possible
				  by simply writing the matrix to file using the :class:`Map class <pycoalescence.map.Map>`.

		.. note:: The generated density map will have dimensions 1 by xy (where x, y are the dimensions of the original
				  density matrix. However, the dispersal matrix should still be compatible with the original density
				  matrix as a x by y tif file.

		:param density_matrix: a numpy matrix containing the density probabilities
		:param dispersal_matrix: a numpy matrix containing the dispersal probabilities
		"""
		if not isinstance(density_matrix, np.ndarray) or not isinstance(dispersal_matrix, np.ndarray):
			raise TypeError("Supplied matrices must be numpy arrays.")
		x_dim, y_dim = dispersal_matrix.shape
		x_density, y_density = density_matrix.shape
		if not x_density * y_density == x_dim or x_dim != y_dim:
			raise ValueError("Density matrix should have dimensions x by y and dispersal matrix xy by xy.")
		# create the patches first
		for y in range(y_dim):
			src_x, src_y = convert_index_to_x_y(y, x_density)
			src_density = density_matrix[src_y, src_x]
			self.add_patch(y, src_density, dispersal_matrix[y, y])
		for y in range(y_dim):
			for x in range(x_dim):
				if x != y:
					self.add_dispersal(y, x, dispersal_matrix[y, x])
		self.generate_files()


def convert_index_to_x_y(index, dim):
	"""
	Converts an index to an x, y coordinate.

	Used when mapping from 1-D space to 2-D space.

	:param index: the index to convert from
	:param dim: the x dimension of the matrix

	:return: a tuple of integers containing the x and y coordinates
	:rtype: tuple
	"""
	density_x = index % dim
	density_y = math.floor(index / dim)
	return int(density_x), int(density_y)



"""
This module contains the C++ shared object files which are accessed via python.
"""
from __future__ import absolute_import

try:
	try:
		from . import libnecsim
	except ImportError:
		import libnecsim
	necsimError = libnecsim.necsimError
except ImportError as ie:
	import logging

	logging.warning("Cannot import libnecsim: {}".format(ie))
	libnecsim = None
	necsimError = None



"""
Calculates landscape-level metrics, including mean distance to nearest-neighbour for each habitat cell and clumpiness.
"""
import logging
try:
	from .necsim import libnecsim
	from .map import Map
	from .system_operations import write_to_log
except ImportError:  # pragma: no cover
	from necsim import libnecsim
	from map import Map
	from system_operations import write_to_log


class LandscapeMetrics(Map):
	"""
	Calculates the mean nearest-neighbour for cells across a landscape. See :ref:`here <landscape_metrics>` for details.
	"""

	def __init__(self, file=None, logging_level=logging.WARNING):
		"""
		Initialises the loggers and default map variables.

		:param file: the path to the map file
		:param logging_level: the logging level to report at
		"""
		Map.__init__(self, file)
		self.logger = logging.Logger("pycoalescence.landscapemetric")
		self._create_logger(logging_level=logging_level)
		self.c_LM_calc = libnecsim.CLandscapeMetricsCalculator(self.logger, write_to_log)

	def __del__(self):
		"""Safely destroy the C++ objects."""
		self.c_LM_calc = None
		Map.__del__(self)

	def get_mnn(self):
		"""
		Calculates the mean nearest-neighbour for cells across a landscape. See :ref:`here <landscape_metrics_mnn>` for
		details.

		:return: the mean distance to the nearest neighbour of a cell.

		:rtype: float
		"""
		self.c_LM_calc.import_map(self.file_name)
		return self.c_LM_calc.calculate_MNN()

	def get_clumpiness(self):
		"""
		Calculates the clumpiness metric for the landscape, a measure of how spread out the points are across the
		landscape. See :ref:`here <landscape_metrics_clumpy>` for details.

		:return: the CLUMPY metric

		:rtype: float
		"""
		self.c_LM_calc.import_map(self.file_name)
		return self.c_LM_calc.calculate_CLUMPY()



"""
Combine simulation outputs from separate guilds. Detailed :ref:`here <merging_simulations>`.

:py:class:`~Merger` will output a single database file, merging the various biodiversity tables into one.

Metrics are also calculated for the entire system, with a guild reference of 0.

All standard routines provided in :py:class:`~pycoalescence.coal_analyse.CoalescenceTree` can then be performed on the
combined database.
"""
from __future__ import absolute_import

import logging
import os

try:
	import sqlite3
except ImportError:  # pragma: no cover
	# Python 3 compatibility
	import sqlite as sqlite3
try:
	from math import isclose
except ImportError:  # pragma: no cover
	from .system_operations import isclose

from .coalescence_tree import CoalescenceTree
from .sqlite_connection import check_sql_table_exist, fetch_table_from_sql
from .system_operations import elegant_pairing, check_parent


class Merger(CoalescenceTree):
	"""
	Merges simulation outputs into a single database. Inherits from
	:py:class:`~pycoalescence.coal_analyse.CoalescenceTree` to provide all routines in the same object.
	"""

	def __init__(self, database=None, logging_level=logging.WARNING, log_output=None, expected=False):
		"""
		Sets up the Merger object.

		:param database: the path to the Merger database object
		:param logging_level: the logging level to assume
		:param log_output: an output stream or file to output logs to
		:param expected: if true, the output database is expected to exist
		"""
		super(Merger, self).__init__(logging_level=logging_level, log_output=log_output)
		self.simulation_list = {}
		if database is not None:
			self.set_database(filename=database, expected=expected)
		self.species_list_parent_counter = 0
		self.parameters_list = []
		self.guild = 0
		self.species_list = []
		self.species_locations = []
		self.species_abundances = []
		self.fragment_abundances = []
		self.species_richness = []
		self.fragment_richness = []
		self.fragment_octaves = []
		self.community_parameters = []
		self.metacommunity_parameters = []
		self.max_species_id = 0
		self.max_locations_id = 0

	def __del__(self):
		"""
		Closes the connection to the database.
		"""
		if self.database:
			self._close_database()

	def _open_database(self, filename):
		"""
		Opens a connection to the database.

		:param filename: the path to the db file to open

		:rtype: None
		"""
		self.file = filename
		check_parent(filename)
		try:
			self.database = sqlite3.connect(filename)
		except sqlite3.Error as e:  # pragma: no cover
			try:
				self.database.close()
			except AttributeError:
				pass
			self.database = None
			raise IOError("Error opening SQLite database: " + str(e))

	def _close_database(self):
		"""Closes the database connection."""
		if self.database is not None:
			self.database.close()
			self.database = None
			self.cursor = None

	def set_database(self, filename, expected=False):
		"""
		Sets the output database for the merged simulations

		Assumes no database currently exists, and will create one.

		:raises IOError: if the output database already exists

		:param filename: the filename to output merged simulations into
		:param expected: if true, expects the output to exist

		:rtype: None
		"""
		if os.path.exists(filename) and not expected:
			raise IOError("Database already exists at {}".format(filename))
		self._open_database(filename)
		if not expected:
			self._create_simulation_parameters()
			self._create_species_list()
			_ = self.get_added_simulations()

	def _create_simulation_parameters(self):
		"""
		Creates the SIMULATION_PARAMETERS table in the database.

		:raises IOError: if the database has not been successfully created
		"""
		if self.database is None:
			raise IOError("Cannot create a SIMULATION_PARAMETERS table as the database has not been created.")
		self.cursor = self.database.cursor()
		create_sql = "CREATE TABLE SIMULATION_PARAMETERS (seed INT not null, job_type INT NOT NULL, " \
					 "output_dir TEXT NOT NULL, speciation_rate DOUBLE NOT NULL, sigma DOUBLE NOT NULL,tau DOUBLE NOT " \
					 "NULL, deme INT NOT NULL, sample_size DOUBLE NOT NULL, max_time INT NOT NULL, " \
					 "dispersal_relative_cost DOUBLE NOT NULL, min_num_species INT NOT NULL, " \
					 "habitat_change_rate DOUBLE NOT NULL, gen_since_historical DOUBLE NOT NULL, " \
					 "time_config_file TEXT NOT NULL, coarse_map_file TEXT NOT NULL, coarse_map_x INT NOT NULL, " \
					 "coarse_map_y INT NOT NULL, coarse_map_x_offset INT NOT NULL, coarse_map_y_offset INT NOT NULL, " \
					 "coarse_map_scale DOUBLE NOT NULL, fine_map_file TEXT NOT NULL, fine_map_x INT NOT NULL, " \
					 "fine_map_y INT NOT NULL, fine_map_x_offset INT NOT NULL, fine_map_y_offset INT NOT NULL, " \
					 "sample_file TEXT NOT NULL, grid_x INT NOT NULL, grid_y INT NOT NULL, sample_x INT NOT NULL, " \
					 "sample_y INT NOT NULL, sample_x_offset INT NOT NULL, sample_y_offset INT NOT NULL, " \
					 "historical_coarse_map TEXT NOT NULL, historical_fine_map TEXT NOT NULL, sim_complete INT NOT NULL, " \
					 "dispersal_method TEXT NOT NULL, m_probability DOUBLE NOT NULL, cutoff DOUBLE NOT NULL, " \
					 "restrict_self INT NOT NULL, landscape_type TEXT NOT NULL, protracted INT NOT NULL, " \
					 "min_speciation_gen DOUBLE NOT NULL, max_speciation_gen DOUBLE NOT NULL, " \
					 "dispersal_map TEXT NOT NULL, guild INT PRIMARY KEY NOT NULL, filename TEXT NOT NULL);"
		try:
			self.cursor.execute(create_sql)
			self.database.commit()
		except sqlite3.Error as soe:  # pragma: no cover
			raise IOError("Could not create SIMULATION_PARAMETERS table in database {}: {}".format(self.file, soe))

	def _create_community_parameters(self):
		"""
		Creates the COMMUNITY_PARAMETERS table in the database.

		:raises IOError: if the database has not been successfully created
		"""
		if self.database is None:
			raise IOError("Cannot create a COMMUNITY_PARAMETERS table as the database has not been created.")
		self.cursor = self.database.cursor()
		create_sql = "CREATE TABLE IF NOT EXISTS COMMUNITY_PARAMETERS (reference INT PRIMARY KEY NOT NULL," \
					 " speciation_rate DOUBLE NOT NULL, time DOUBLE NOT NULL, fragments INT NOT NULL, " \
					 "metacommunity_reference INT);"
		try:
			self.cursor.execute(create_sql)
			self.database.commit()
		except sqlite3.Error as soe:  # pragma: no cover
			raise IOError("Could not create COMMUNITY_PARAMETERS table in database {}: {}".format(self.file, soe))

	def _create_metacommunity_parameters(self):
		"""
		Creates the METACOMMUNITY_PARAMETERS table in the database.

		:raises IOError: if the database has not been successfully created
		"""
		if self.database is None:
			raise IOError("Cannot create a METACOMMUNITY_PARAMETERS table as the database has not been created.")
		self.cursor = self.database.cursor()
		create_sql = "CREATE TABLE IF NOT EXISTS METACOMMUNITY_PARAMETERS (reference INT PRIMARY KEY NOT NULL," \
					 " speciation_rate DOUBLE NOT NULL, metacommunity_size DOUBLE NOT NULL, option TEXT NOT NULL," \
					 "external_reference INT NOT NULL);"
		try:
			self.cursor.execute(create_sql)
			self.database.commit()
		except sqlite3.Error as soe:  # pragma: no cover
			raise IOError("Could not create COMMUNITY_PARAMETERS table in database {}: {}".format(self.file, soe))

	def _create_species_list(self):
		"""
		Creates the SPECIES_LIST table in the database.

		:raises IOError: if the database has not been successfully created
		"""
		if self.database is None:
			raise IOError("Cannot create a SPECIES_LIST table as the database has not been created.")
		self.cursor = self.database.cursor()
		create_sql = "CREATE TABLE IF NOT EXISTS SPECIES_LIST (ID int PRIMARY KEY NOT NULL, unique_spec INT NOT NULL, " \
					 "xval INT NOT NULL, yval INT NOT NULL, xwrap INT NOT NULL, ywrap INT NOT NULL, tip INT NOT NULL, " \
					 "speciated INT NOT NULL, " \
					 "parent INT NOT NULL, existence INT NOT NULL, randnum DOUBLE NOT NULL, gen_alive INT NOT NULL, " \
					 "gen_added DOUBLE NOT NULL, guild INT NOT NULL);"
		try:
			self.cursor.execute(create_sql)
			self.database.commit()
		except sqlite3.Error as soe:  # pragma: no cover
			raise IOError("Could not create SPECIES_LIST table in database {}: {}".format(self.file, soe))

	def _create_species_locations(self):
		"""
		Creates the SPECIES_LOCATIONS table in the database.

		:raises IOError: if the database has not been successfully created
		"""
		if self.database is None:
			raise IOError("Cannot create a SPECIES_LOCATIONS table as the database has not been created.")
		self.cursor = self.database.cursor()
		create_sql = "CREATE TABLE IF NOT EXISTS SPECIES_LOCATIONS (ID int PRIMARY KEY NOT NULL, species_id INT " \
					 "NOT NULL, x INT NOT NULL, y INT NOT NULL, community_reference INT NOT NULL, guild INT NOT NULL);"
		try:
			self.cursor.execute(create_sql)
			self.database.commit()
		except sqlite3.Error as soe:  # pragma: no cover
			raise IOError("Could not create SPECIES_LOCATIONS table in database {}: {}".format(self.file, soe))

	def _create_species_abundances(self):
		"""
		Creates the SPECIES_ABUNDANCES table in the database.

		:raises IOError: if the database has not been successfully created
		"""
		if self.database is None:
			raise IOError("Cannot create a SPECIES_ABUNDANCES table as the database has not been created.")
		self.cursor = self.database.cursor()
		create_sql = "CREATE TABLE IF NOT EXISTS SPECIES_ABUNDANCES (ID int PRIMARY KEY NOT NULL, " \
					 "species_id INT NOT NULL, no_individuals INT NOT NULL, community_reference INT NOT NULL, " \
					 "guild INT NOT NULL);"
		try:
			self.cursor.execute(create_sql)
			self.database.commit()
		except sqlite3.Error as soe:  # pragma: no cover
			raise IOError("Could not create SPECIES_ABUNDANCES table in database {}: {}".format(self.file, soe))

	def _create_fragment_abundances(self):
		"""
		Creates the FRAGMENT_ABUNDANCES table in the database.

		:raises IOError: if the database has not been successfully created
		"""
		if self.database is None:
			raise IOError("Cannot create a FRAGMENT_ABUNDANCES table as the database has not been created.")
		self.cursor = self.database.cursor()
		create_sql = "CREATE TABLE IF NOT EXISTS FRAGMENT_ABUNDANCES (ID int PRIMARY KEY NOT NULL," \
					 " fragment TEXT NOT NULL, area DOUBLE NOT NULL, size INT NOT NULL,  species_id INT NOT NULL, " \
					 "no_individuals INT NOT NULL, community_reference int NOT NULL," \
					 " guild INT NOT NULL);"
		try:
			self.cursor.execute(create_sql)
			self.database.commit()
		except sqlite3.Error as soe:  # pragma: no cover
			raise IOError("Could not create SPECIES_ABUNDANCES table in database {}: {}".format(self.file, soe))

	def _create_fragment_octaves(self, guilds=False):
		"""
		Creates the FRAGMENT_OCTAVES table in the database.

		:param guilds: if true, creates a separate table for individual guilds, rather than for the whole community

		:raises IOError: if the database has not been successfully created
		"""
		if self.database is None:
			raise IOError("Cannot create a FRAGMENT_OCTAVES table as the database has not been created.")
		if guilds:
			additional1 = "_GUILDS"
			additional2 = ", guild INT NOT NULL"
		else:
			additional1 = ""
			additional2 = ""
		self.cursor = self.database.cursor()
		create_sql = "CREATE TABLE IF NOT EXISTS FRAGMENT_OCTAVES{} (ref INT PRIMARY KEY NOT NULL, fragment TEXT NOT NULL, " \
					 "octave INT NOT NULL, richness INT NOT NULL{});".format(additional1, additional2)
		try:
			self.cursor.execute(create_sql)
			self.database.commit()
		except sqlite3.Error as soe:  # pragma: no cover
			raise IOError("Could not create SPECIES_ABUNDANCES table in database {}: {}".format(self.file, soe))

	def _create_species_richness(self, guilds=False):
		"""
		Creates the SPECIES_RICHNESS table in the database.

		:raises IOError: if the database has not been successfully created
		"""
		if self.database is None:
			raise IOError("Cannot create a SPECIES_RICHNESS table as the database has not been created.")
		if guilds:
			additional1 = "_GUILDS"
			additional2 = ", guild INT NOT NULL"
		else:
			additional1 = ""
			additional2 = ""
		self.cursor = self.database.cursor()
		create_sql = "CREATE TABLE IF NOT EXISTS SPECIES_RICHNESS{} (ref INT PRIMARY KEY NOT NULL," \
					 " community_reference INT NOT NULL, richness INT NOT NULL{})".format(additional1, additional2)

		try:
			self.cursor.execute(create_sql)
			self.database.commit()
		except sqlite3.Error as soe:  # pragma: no cover
			raise IOError("Could not create SPECIES_RICHNESS table in database {}: {}".format(self.file, soe))

	def _create_fragment_richness(self, guilds=False):
		"""
		Creates the FRAGMENT_RICHNESS table in the database.

		:param guilds: if true, creates a separate table for individual guilds, rather than for the whole community

		:raises IOError: if the database has not been successfully created
		"""
		if self.database is None:
			raise IOError("Cannot create a FRAGMENT_RICHNESS table as the database has not been created.")
		if guilds:
			additional1 = "_GUILDS"
			additional2 = ", guild INT NOT NULL"
		else:
			additional1 = ""
			additional2 = ""
		self.cursor = self.database.cursor()
		create_sql = "CREATE TABLE IF NOT EXISTS FRAGMENT_RICHNESS{} (ref INT PRIMARY KEY NOT NULL, fragment TEXT NOT NULL," \
					 " community_reference INT NOT NULL,  richness INT NOT NULL{})".format(additional1, additional2)
		try:
			self.cursor.execute(create_sql)
			self.database.commit()
		except sqlite3.Error as soe:  # pragma: no cover
			raise IOError("Could not create FRAGMENT_RICHNESS table in database {}: {}".format(self.file, soe))

	def _read_simulation_parameters(self, input_simulation):
		"""
		Reads the simulation parameters from the input file and returns a list containing these parameters

		:param input_simulation: the simulation to read the simulation parameters from

		:return: list containing the simulation parameters from the input file
		"""
		output = fetch_table_from_sql(input_simulation, "SIMULATION_PARAMETERS")
		if len(output[0]) != 44:  # pragma: no cover
			raise IOError("SIMULATION_PARAMETERS table does not contain 44 columns, contains {}."
						  " Check database is correct format.".format(len(output[0])))
		return output[0]

	def _add_simulation_parameters(self, parameters_list, guild, filename):
		"""
		Adds the provided parameters to the output statement for SIMULATION_PARAMETERS with the provided guild reference.
		:param list parameters_list: a list containing the SIMULATION_PARAMETERS outputs.
		:param int guild: the guild reference number for
		:param str filename: the simulation file associated with these parameters
		:return: None
		:rtype: None
		"""
		parameters_list.extend([guild, filename])
		self.parameters_list.append(tuple(parameters_list))

	def _write_simulation_parameters(self):
		"""
		Writes out the parameters_list object to the output database.

		:return: None
		:rtype: None
		"""
		if self.database is None:
			raise IOError("Cannot insert to SPECIES_PARAMETERS table as the database has not been created.")
		self.cursor = self.database.cursor()
		create_sql = "INSERT INTO SIMULATION_PARAMETERS VALUES (?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?," \
					 "?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?);"
		try:
			self.cursor.executemany(create_sql, self.parameters_list)
			self.database.commit()
		except sqlite3.Error as soe:  # pragma: no cover
			raise IOError("Could not create SIMULATION_PARAMETERS table in database {}: {}".format(self.file, soe))

	def _read_species_list(self, input_simulation):
		"""
		Reads the SPECIES_LIST table from the input simulation
		:param input_simulation: the completed simulation to read the SPECIES_LIST object from.
		:return:
		"""
		return fetch_table_from_sql(input_simulation, "SPECIES_LIST")

	def _add_species_list(self, species_list, guild):
		"""
		Adds to the species_list object with the provided list of species and guild reference.
		:param species_list: the list of lists of species information to append
		:param guild: the guild to append to each species row.
		"""
		for i, row in enumerate(species_list):
			row[0] = int(elegant_pairing(row[0], guild))
			if i != 0 and row[8] != 0:
				row[8] = row[8] + self.species_list_parent_counter
			row.append(guild)
		self.species_list_parent_counter += len(species_list)
		self.species_list.extend([tuple(x) for x in species_list])

	def _write_species_list(self):
		"""
		Writes out the species_list object to the output database.

		:return: None
		:rtype: None
		"""
		if self.database is None:
			raise IOError("Cannot insert to SPECIES_LIST table as the database has not been created.")
		self.cursor = self.database.cursor()
		create_sql = "INSERT INTO SPECIES_LIST VALUES (?,?,?,?,?,?,?,?,?,?,?,?,?,?);"
		try:
			self.cursor.executemany(create_sql, self.species_list)
			self.database.commit()
		except sqlite3.Error as soe:  # pragma: no cover
			raise IOError("Could not insert into SPECIES_LIST table in database {}: {}".format(self.file, soe))

	def _read_species_locations(self, input_simulation):
		"""
		Reads the SPECIES_LOCATIONS table from the input simulation
		:param input_simulation: the completed simulation to read the SPECIES_LIST object from.
		:return:
		"""
		return fetch_table_from_sql(input_simulation, "SPECIES_LOCATIONS")

	def _add_species_locations(self, species_locations, guild):
		"""
		Adds to the species_locations object with the provided list of species and guild reference.
		:param species_list: the list of lists of species information to append
		:param guild: the guild to append to each species row.
		"""
		tmp_max = self.max_locations_id
		for row in species_locations:
			row[0] = elegant_pairing(row[0], guild)
			row.append(guild)
			self.max_locations_id = max(row[1], self.max_locations_id)
			row[1] += tmp_max
		self.species_locations.extend([tuple(x) for x in species_locations])

	def _write_species_locations(self):
		"""
		Writes out the species_locations object to the output database.

		:return: None
		:rtype: None
		"""
		if self.database is None:
			raise IOError("Cannot insert to SPECIES_LOCATIONS table as the database has not been created.")
		self.cursor = self.database.cursor()
		create_sql = "INSERT INTO SPECIES_LOCATIONS VALUES (?,?,?,?,?,?);"
		try:
			self.cursor.executemany(create_sql, self.species_locations)
			self.database.commit()
		except sqlite3.Error as soe:  # pragma: no cover
			raise IOError("Could not insert into SPECIES_LOCATIONS table in database {}: {}".format(self.file, soe))

	def _read_species_abundances(self, input_simulation):
		"""
		Reads the SPECIES_ABUNDANCES table from the input simulation.

		:param input_simulation: the completed simulation to read the SPECIES_ABUNDANCES object from.
		:return:
		"""
		return fetch_table_from_sql(input_simulation, "SPECIES_ABUNDANCES")

	def _add_species_abundances(self, species_abundances, guild):
		"""
		Adds to the species_abundances object with the provided list of species and guild reference.

		:param species_abundances: the list of lists of species information to append
		:param guild: the guild to append to each species row.
		"""
		tmp_max = self.max_species_id
		for row in species_abundances:
			row[0] = elegant_pairing(row[0], guild)
			row.append(guild)
			self.max_species_id = max(row[1], self.max_species_id)
			row[1] += tmp_max
		self.species_abundances.extend([tuple(x) for x in species_abundances])

	def _write_species_abundances(self):
		"""
		Writes out the species_list object to the output database.

		:return: None
		:rtype: None
		"""
		if self.database is None:
			raise IOError("Cannot insert to SPECIES_ABUNDANCES table as the database has not been created.")
		self.cursor = self.database.cursor()
		create_sql = "INSERT INTO SPECIES_ABUNDANCES VALUES (?,?,?,?,?);"
		try:
			self.cursor.executemany(create_sql, self.species_abundances)
			self.database.commit()
		except sqlite3.Error as soe:  # pragma: no cover
			raise IOError("Could not insert into SPECIES_ABUNDANCES table in database {}: {}".format(self.file, soe))

	def _read_fragment_abundances(self, input_simulation):
		"""
		Reads the FRAGMENT_ABUNDANCES table from the input simulation
		:param input_simulation: the completed simulation to read the FRAGMENT_ABUNDANCES object from.
		:return:
		"""
		return fetch_table_from_sql(input_simulation, "FRAGMENT_ABUNDANCES")

	def _add_fragment_abundances(self, fragment_abundances, guild):
		"""
		Adds to the fragment_abundances object with the provided list of species and guild reference.
		:param fragment_abundances: the list of lists of species information to append
		:param guild: the guild to append to each species row.
		"""
		for row in fragment_abundances:
			row[0] = elegant_pairing(row[0], guild)
			row.append(guild)
		self.fragment_abundances.extend([tuple(x) for x in fragment_abundances])

	def _write_fragment_abundances(self):
		"""
		Writes out the fragment_abundances object to the output database.

		:return: None
		:rtype: None
		"""
		if self.database is None:
			raise IOError("Cannot insert to FRAGMENT_ABUNDANCES table as the database has not been created.")
		self.cursor = self.database.cursor()
		create_sql = "INSERT INTO FRAGMENT_ABUNDANCES VALUES (?, ?, ?, ?, ?, ?, ?, ?);"
		try:
			self.cursor.executemany(create_sql, self.fragment_abundances)
			self.database.commit()
		except sqlite3.Error as soe:  # pragma: no cover
			raise IOError("Could not insert into FRAGMENT_ABUNDANCES table in database {}: {}".format(self.file, soe))

	def _read_species_richness(self, input_simulation):
		"""
		Reads the SPECIES_RICHNESS table from the input simulation

		:param input_simulation: the completed simulation to read the SPECIES_RICHNESS object from.
		:return:
		"""
		return fetch_table_from_sql(input_simulation, "SPECIES_RICHNESS")

	def _add_species_richness(self, species_richness, guild):
		"""
		Adds to the species_richness object with the provided list of species and guild reference.
		:param species_richness: the list of lists of species information to append
		:param guild: the guild to append to each species row.
		"""
		for row in species_richness:
			row[0] = elegant_pairing(row[0], guild)
			row.append(guild)
		self.species_richness.extend([tuple(x) for x in species_richness])

	def _write_species_richness(self, guilds=False):
		"""
		Writes out the species_richness object to the output database.

		:param guilds: if true, creates a separate table for individual guilds, rather than for the whole community

		:return: None
		:rtype: None
		"""
		if self.database is None:
			raise IOError("Cannot insert to SPECIES_RICHNESS table as the database has not been created.")
		if guilds:
			additional = "_GUILDS VALUES (?,?,"
		else:
			additional = " VALUES (?,"
		self.cursor = self.database.cursor()
		create_sql = "INSERT INTO SPECIES_RICHNESS{} ?,?);".format(additional)
		try:
			self.cursor.executemany(create_sql, self.species_richness)
			self.database.commit()
		except sqlite3.Error as soe:  # pragma: no cover
			raise IOError("Could not insert into SPECIES_RICHNESS table in database {}: {}".format(self.file, soe))

	def _read_fragment_richness(self, input_simulation):
		"""
		Reads the FRAGMENT_RICHNESS table from the input simulation

		:param input_simulation: the completed simulation to read the FRAGMENT_RICHNESS object from.
		:return:
		"""
		return fetch_table_from_sql(input_simulation, "FRAGMENT_RICHNESS")

	def _add_fragment_richness(self, fragment_richness, guild):
		"""
		Adds to the species_richness object with the provided list of species and guild reference.
		:param fragment_richness: the list of lists of species information to append
		:param guild: the guild to append to each species row.
		"""
		for row in fragment_richness:
			row[0] = elegant_pairing(row[0], guild)
			row.append(guild)
		self.fragment_richness.extend([tuple(x) for x in fragment_richness])

	def _write_fragment_richness(self, guilds=False):
		"""
		Writes out the fragment_richness object to the output database.

		:param guilds: if true, creates a separate table for individual guilds, rather than for the whole community


		:return: None
		:rtype: None
		"""
		if self.database is None:
			raise IOError("Cannot insert to FRAGMENT_RICHNESS table as the database has not been created.")
		if guilds:
			additional = "_GUILDS VALUES (?,"
		else:
			additional = " VALUES ("
		self.cursor = self.database.cursor()
		create_sql = "INSERT INTO FRAGMENT_RICHNESS{} ?,?,?,?);".format(additional)
		try:
			self.cursor.executemany(create_sql, self.fragment_richness)
			self.database.commit()
		except sqlite3.Error as soe:  # pragma: no cover
			raise IOError("Could not insert into FRAGMENT_RICHNESS table in database {}: {}".format(self.file, soe))

	def _read_fragment_octaves(self, input_simulation):
		"""
		Reads the FRAGMENT_ABUNDANCES table from the input simulation

		:param input_simulation: the completed simulation to read the FRAGMENT_ABUNDANCES object from.
		:return:
		"""
		try:
			db = sqlite3.connect(input_simulation)
			c = db.cursor()
			c.execute("SELECT ref, fragment, octave, richness FROM FRAGMENT_OCTAVES")
			out = [list(x) for x in c.fetchall()]
			db.close()
			db = None
			return out
		except sqlite3.Error as soe:  # pragma: no cover
			raise IOError("Cannot fetch FRAGMENT_OCTAVES from database {}: {}".format(input_simulation, soe))

	def _add_fragment_octaves(self, fragment_octaves, guild):
		"""
		Adds to the fragment_octaves object with the provided list of species and guild reference.
		:param fragment_octaves: the list of lists of species information to append
		:param guild: the guild to append to each species row.
		"""
		for row in fragment_octaves:
			row[0] = elegant_pairing(row[0], guild)
			row.append(guild)
		self.fragment_octaves.extend([tuple(x) for x in fragment_octaves])

	def _write_fragment_octaves(self, guilds=False):
		"""
		Writes out the fragment_octaves object to the output database.

		:param guilds: if true, creates a separate table for individual guilds, rather than for the whole community

		:return: None
		:rtype: None
		"""
		if self.database is None:
			raise IOError("Cannot insert to FRAGMENT_ABUNDANCES table as the database has not been created.")
		if guilds:
			additional = "_GUILDS VALUES (?,"
		else:
			additional = " VALUES ("
		self.cursor = self.database.cursor()
		create_sql = "INSERT INTO FRAGMENT_OCTAVES{} ?,?,?,?);".format(additional)
		try:
			self.cursor.executemany(create_sql, self.fragment_octaves)
			self.database.commit()
		except sqlite3.Error as soe:  # pragma: no cover
			raise IOError("Could not insert into FRAGMENT_ABUNDANCES table in database {}: {}".format(self.file, soe))

	def _add_community_parameters(self, input_simulation):
		"""
		Adds the community parameters to the internal object, checking to ensure that no repeat references exist with
		different parameters.

		:param input_simulation: completed simulation to obtain community references from

		:raises ValueError: if there is a parameter mismatch between the references
		"""
		community_parameters = fetch_table_from_sql(input_simulation, "COMMUNITY_PARAMETERS")
		if len(self.community_parameters) > 0:
			for reference, speciation_rate, time, fragments, metacommunity_reference in community_parameters:
				subsetted_params = [x for x in self.community_parameters if
									isclose(speciation_rate, x[1]) and isclose(time, x[2], abs_tol=0.00001) and
									metacommunity_reference == x[4]]
				if len(subsetted_params) > 0:
					if subsetted_params[0][0] != reference:  # pragma: no cover
						raise ValueError("Parameter mismatch in references for speciation rate = {} and"
										 " time = {}".format(speciation_rate, time))
				else:  # pragma: no cover
					self.logger.error("Parameter set for reference of {} "
									  "does not exist in all simulations.".format(reference))
					self.community_parameters.append([reference, speciation_rate, time, fragments,
													  metacommunity_reference])
			for reference, speciation_rate, time, fragments, metacommunity_reference in self.community_parameters:
				subsetted_params = [x for x in community_parameters if isclose(speciation_rate, x[1]) and
									isclose(time, x[2], abs_tol=0.00001) and metacommunity_reference == x[4]]
				if len(subsetted_params) == 0:  # pragma: no cover
					self.logger.error("Parameter set for reference of {} "
									  "does not exist in all simulations.".format(reference))
					self.community_parameters.append([reference, speciation_rate, time, fragments,
													  metacommunity_reference])
		else:
			self.community_parameters.extend(community_parameters)

	def _write_community_parameters(self):
		"""
		Outputs the community parameters into the database
		"""
		if self.database is None:
			raise IOError("Cannot insert to COMMUNITY_PARAMETERS table as the database has not been created.")
		self.cursor = self.database.cursor()
		create_sql = "INSERT INTO COMMUNITY_PARAMETERS VALUES (?,?,?,?,?);"
		try:
			self.cursor.executemany(create_sql, self.community_parameters)
			self.database.commit()
		except sqlite3.Error as soe:  # pragma: no cover
			raise IOError("Could not insert into COMMUNITY_PARAMETERS table in database {}: {}".format(self.file, soe))

	def _add_metacommunity_parameters(self, input_simulation):
		"""
		Adds the metacommunity parameters to the internal object, checking to ensure that no repeat references exist
		with different parameters.

		:param input_simulation: completed simulation to obtain metacommunity references from

		:raises ValueError: if there is a parameter mismatch between the references
		"""
		metacommunity_parameters = fetch_table_from_sql(input_simulation, "METACOMMUNITY_PARAMETERS")
		if len(self.metacommunity_parameters) > 0:
			for reference, speciation_rate, metacommunity_size, option, external_reference in metacommunity_parameters:
				subsetted_params = [x for x in self.metacommunity_parameters if isclose(speciation_rate, x[1]) and
									isclose(metacommunity_size, x[2]) and option == x[3] and external_reference == x[4]]
				if len(subsetted_params) > 0 and len(self.metacommunity_parameters) > 0:
					if subsetted_params[0][0] != reference:  # pragma: no cover
						raise ValueError("Parameter mismatch in metacommunity references for speciation rate = {} and"
										 " size = {}".format(speciation_rate, metacommunity_size))
				else:  # pragma: no cover
					self.logger.error("Parameter set for metacommunity reference of {} "
									  "(speciation rate={}, metacommunity_size={}, option={}, external_reference={}) "
									  "does not exist in all simulations.".format(reference, speciation_rate,
																				  metacommunity_size, option,
																				  external_reference))
					self.metacommunity_parameters.append([reference, speciation_rate, metacommunity_size, option,
														  external_reference])
			for reference, speciation_rate, metacommunity_size, \
				option, external_reference in self.metacommunity_parameters:
				subsetted_params = [x for x in metacommunity_parameters if isclose(speciation_rate, x[1]) and
									metacommunity_size == x[2] and option == x[3] and external_reference == x[4]]
				if len(subsetted_params) == 0:  # pragma: no cover
					self.logger.error("Parameter set for metacommunity reference of {} "
									  "(speciation rate={}, metacommunity_size={}, option={}, external_reference={}) "
									  "does not exist in added simulation.".format(reference, speciation_rate,
																				   metacommunity_size, option,
																				   external_reference))
					self.metacommunity_parameters.append([reference, speciation_rate, metacommunity_size, option,
														  external_reference])

		else:
			self.metacommunity_parameters.extend(metacommunity_parameters)

	def _write_metacommunity_parameters(self):
		if self.database is None:
			raise IOError("Cannot insert to METACOMMUNITY_PARAMETERS table as the database has not been created.")
		self.cursor = self.database.cursor()
		create_sql = "INSERT INTO METACOMMUNITY_PARAMETERS VALUES (?,?,?, ?, ?);"
		try:
			self.cursor.executemany(create_sql, self.metacommunity_parameters)
			self.database.commit()
		except sqlite3.Error as soe:  # pragma: no cover
			raise IOError(
				"Could not insert into METACOMMUNITY_PARAMETERS table in database {}: {}".format(self.file, soe))

	def get_added_simulations(self):
		"""
		Gets the simulations which have already been added to the database.

		:return: dictionary of simulations and guild numbers
		:rtype: dict
		"""
		simulations = self.cursor.execute("SELECT filename, guild FROM SIMULATION_PARAMETERS").fetchall()
		for file_name, guild in simulations:
			if file_name in self.simulation_list.keys():
				if guild != self.simulation_list[file_name]:
					raise ValueError("Simulation from {} does not match the "
									 "recorded guild ({} != {}).".format(file_name, guild,
																		 self.simulation_list[file_name]))
			else:
				self.simulation_list[file_name] = guild
		return self.simulation_list

	def add_simulation(self, input_simulation):
		"""
		Adds a simulation to the list of merged simulations.

		This also calls the relevant merges for the tables that exist in the provided database.

		:param input_simulation: either the path to the input simulation, a Coalescence class object, or a CoalescenceTree object
			which contains the completed simulation.

		:return: None

		:rtype: None
		"""
		if isinstance(input_simulation, CoalescenceTree):
			t = input_simulation
		else:
			t = CoalescenceTree(input_simulation)
		simulation_path = t.file
		_ = self.get_added_simulations()
		if simulation_path in self.simulation_list.keys():
			raise ValueError("Simulation path of {} has already been added to Merger.".format(simulation_path))
		self.guild += 1
		self.simulation_list[simulation_path] = self.guild
		# Ensure that any open sql connections are closed
		del t
		# First add the parameters (which might throw some simple errors)
		self._add_simulation_parameters(self._read_simulation_parameters(simulation_path), self.guild, simulation_path)
		if check_sql_table_exist(simulation_path, "COMMUNITY_PARAMETERS"):
			self._create_community_parameters()
			self._add_community_parameters(simulation_path)
		if check_sql_table_exist(simulation_path, "METACOMMUNITY_PARAMETERS"):
			self._create_metacommunity_parameters()
			self._add_metacommunity_parameters(simulation_path)
		# Now the longer routines
		self._add_species_list(self._read_species_list(simulation_path), self.guild)
		if check_sql_table_exist(simulation_path, "SPECIES_ABUNDANCES"):
			self._create_species_abundances()
			self._add_species_abundances(self._read_species_abundances(simulation_path), self.guild)
		if check_sql_table_exist(simulation_path, "SPECIES_LOCATIONS"):
			self._create_species_locations()
			self._add_species_locations(self._read_species_locations(simulation_path), self.guild)
		if check_sql_table_exist(simulation_path, "FRAGMENT_ABUNDANCES"):
			self._create_fragment_abundances()
			self._add_fragment_abundances(self._read_fragment_abundances(simulation_path), self.guild)
		if check_sql_table_exist(simulation_path, "SPECIES_RICHNESS"):
			self._create_species_richness()
			self._add_species_richness(self._read_species_richness(simulation_path), self.guild)
		if check_sql_table_exist(simulation_path, "FRAGMENT_RICHNESS"):
			self._create_fragment_richness()
			self._add_fragment_richness(self._read_fragment_richness(simulation_path), self.guild)
		if check_sql_table_exist(simulation_path, "FRAGMENT_OCTAVES"):
			self._create_fragment_octaves()
			self._add_fragment_octaves(self._read_fragment_octaves(simulation_path), self.guild)

	def generate_guild_tables(self):
		"""
		Generates a set of tables containing the biodiversity metrics for each guild.

		:rtype: None
		"""
		self.species_richness = []
		self.fragment_richness = []
		self.fragment_octaves = []
		for simulation_path, guild in self.simulation_list.items():
			if check_sql_table_exist(simulation_path, "SPECIES_RICHNESS"):
				self._create_species_richness(guilds=True)
				self._add_species_richness(self._read_species_richness(simulation_path), guild)
			if check_sql_table_exist(simulation_path, "FRAGMENT_RICHNESS"):
				self._create_fragment_richness(guilds=True)
				self._add_fragment_richness(self._read_fragment_richness(simulation_path), guild)
			if check_sql_table_exist(simulation_path, "FRAGMENT_OCTAVES"):
				self._create_fragment_octaves(guilds=True)
				self._add_fragment_octaves(self._read_fragment_octaves(simulation_path), guild)
		if len(self.species_richness) > 0:
			self._write_species_richness(guilds=True)
		if len(self.fragment_richness) > 0:
			self._write_fragment_richness(guilds=True)
		if len(self.fragment_octaves) > 0:
			self._write_fragment_octaves(guilds=True)

	def add_simulations(self, simulation_list):
		"""
		A convenience function that adds each simulation from the list of simulations provided and then writes to the
		database.

		:param simulation_list: list of paths to completed simulations
		"""
		for each in simulation_list:
			self.add_simulation(each)
		self.write()

	def write(self):
		"""
		Writes out all stored simulation parameters to the output database and wipes the in-memory objects.

		This should be called after all simulation have been added, or when RAM usage gets too large for large
		simulations

		"""
		self._write_simulation_parameters()
		if len(self.community_parameters) > 0:
			self._write_community_parameters()
		if len(self.metacommunity_parameters) > 0:
			self._write_metacommunity_parameters()
		self._write_species_list()
		if len(self.species_abundances) > 0:
			self._write_species_abundances()
		if len(self.species_locations) > 0:
			self._write_species_locations()
		if len(self.fragment_abundances) > 0:
			self._write_fragment_abundances()
		if len(self.species_richness) > 0:
			self._create_combined_species_richness()
			self._write_species_richness()
		if len(self.fragment_richness) > 0:
			self._create_combined_fragment_richness()
			self._write_fragment_richness()
		if len(self.fragment_octaves) > 0:
			self._create_combined_fragment_octaves()
			self._write_fragment_octaves()

	def apply(self):
		"""
		Generates the cooalescence tree for the set of speciation parameters.
		This must be run after the main coalescence simulations are complete.
		It will create additional fields and tables in the SQLite database which contains the requested data.
		"""
		self._close_database()
		super(Merger, self).apply()
		self._open_database(self.file)

	def apply_incremental(self):
		"""
		Generates the coalescence tree for the set of speciation parameters. Does not write changes to the database,
		just holds the changes internally.
		"""
		self._close_database()
		super(Merger, self).apply_incremental()
		self._open_database(self.file)

	def output(self):
		"""Outputs the coalescence trees to the same simulation database object."""
		self._close_database()
		super(Merger, self).output()
		self._open_database(self.file)

	def _create_combined_species_richness(self):
		"""
		Combines the species richness guilds so that richness values are for all guilds.
		"""
		if len(self.species_richness) == 0:
			raise ValueError("No species richnesses are detected. Add simulations before combining richness.")
		out_richness = []
		references = set([x[1] for x in self.species_richness])
		i = 0
		for reference in references:
			total = sum([x[2] for x in self.species_richness if x[1] == reference])
			out_richness.append([i, reference, total])
			i += 1
		self.species_richness = out_richness

	def _create_combined_fragment_richness(self):
		"""
		Combines the fragment richness guilds so that richness values are for all guilds in each fragment.
		"""
		if len(self.fragment_richness) == 0:
			raise ValueError("No fragment richnesses are detected. Add simulations before combining richness.")
		out_richness = []
		fragments = set([x[1] for x in self.fragment_richness])
		references = set([x[2] for x in self.fragment_richness])
		i = 0
		for reference in references:
			for fragment in fragments:
				total = sum([x[3] for x in self.fragment_richness if x[1] == fragment and x[2] == reference])
				out_richness.append([i, fragment, reference, total])
				i += 1
		self.fragment_richness = out_richness

	def _create_combined_fragment_octaves(self):
		"""
		Combines the fragment octaves guilds so that octaves values are for all guilds in each fragment.
		"""
		if len(self.fragment_octaves) == 0:
			raise ValueError("No fragment octaves are detected. Add simulations before creating combined octaves")
		octaves = []
		fragments = set([x[1] for x in self.fragment_octaves])
		references = set([x[2] for x in self.fragment_octaves])
		i = 0
		for reference in references:
			for fragment in fragments:
				total = sum([x[3] for x in self.fragment_octaves if x[1] == fragment and x[2] == reference])
				octaves.append([i, fragment, reference, total])
				i += 1
		self.fragment_octaves = octaves



"""
Port older simulation outputs to the updated naming conventions. Should not be required by most users.
"""

import sqlite3
import logging
from pycoalescence.sqlite_connection import check_sql_table_exist


def update_parameter_names(database):
	"""
	Alters the parameters names of SIMULATION_PARAMETERS in the database so that it matches the
	updated naming convention.

	Provided for back-compatibility with older simulations.

	.. note:: If the simulation does not require updating, this function exits silently.

	:param database: the database path to alter the names of

	:return: None

	:rtype: None
	"""
	try:
		db = sqlite3.connect(database)
		c = db.cursor()
		if not check_sql_table_exist(db, "SIMULATION_PARAMETERS"):
			raise IOError("Table SIMULATION_PARAMETERS does not exist in database.")
		# Check if the parameters are already updated
		sql_query = ("SELECT seed, job_type, output_dir, speciation_rate, sigma, tau, deme," 
					" sample_size, max_time, dispersal_relative_cost, min_num_species, habitat_change_rate," 
					" gen_since_historical, time_config_file, coarse_map_file, coarse_map_x, coarse_map_y,"
					" coarse_map_x_offset, coarse_map_y_offset, coarse_map_scale, fine_map_file, fine_map_x, " 
					"fine_map_y, fine_map_x_offset, fine_map_y_offset, sample_file, grid_x, grid_y, sample_x, sample_y,"
					"sample_x_offset, sample_y_offset, historical_coarse_map, historical_fine_map, sim_complete, "
					"dispersal_method, m_probability, cutoff, landscape_type, protracted, "
					"min_speciation_gen, max_speciation_gen, dispersal_map FROM SIMULATION_PARAMETERS")
		try:
			c.execute(sql_query)
			db = None
			return
		except sqlite3.Error:
			pass
		sql_query = "ALTER TABLE SIMULATION_PARAMETERS RENAME TO SIM_P_backup; "
		try:
			c.execute(sql_query)
		except sqlite3.Error:  # pragma: no cover
			c.execute("DROP TABLE SIM_P_backup;")
			c.execute(sql_query)
		sql_query = "CREATE TABLE SIMULATION_PARAMETERS(seed INT PRIMARY KEY not null, job_type INT NOT NULL,"
		sql_query += "output_dir TEXT NOT NULL, speciation_rate DOUBLE NOT NULL, sigma DOUBLE NOT NULL,tau DOUBLE NOT NULL," \
					 " deme INT NOT NULL, "
		sql_query += "sample_size DOUBLE NOT NULL, max_time INT NOT NULL, dispersal_relative_cost DOUBLE NOT NULL, " \
					 "min_num_species ";
		sql_query += "INT NOT NULL, habitat_change_rate DOUBLE NOT NULL, gen_since_historical DOUBLE NOT NULL, "
		sql_query += "time_config_file TEXT NOT NULL, coarse_map_file TEXT NOT NULL, coarse_map_x INT NOT NULL, " \
					 "coarse_map_y INT NOT NULL,"
		sql_query += "coarse_map_x_offset INT NOT NULL, coarse_map_y_offset INT NOT NULL, coarse_map_scale DOUBLE NOT " \
					 "NULL, fine_map_file TEXT NOT NULL, fine_map_x INT NOT NULL,"
		sql_query += "fine_map_y INT NOT NULL, fine_map_x_offset INT NOT NULL, fine_map_y_offset INT NOT NULL, "
		sql_query += "sample_file TEXT NOT NULL, grid_x INT NOT NULL, grid_y INT NOT NULL, sample_x INT NOT NULL, "
		sql_query += "sample_y INT NOT NULL, sample_x_offset INT NOT NULL, sample_y_offset INT NOT NULL, "
		sql_query += "historical_coarse_map TEXT NOT NULL, historical_fine_map TEXT NOT NULL, sim_complete INT NOT NULL, "
		sql_query += "dispersal_method TEXT NOT NULL, m_probability DOUBLE NOT NULL, cutoff DOUBLE NOT NULL, "
		sql_query += "restrict_self INT NOT NULL, landscape_type TEXT NOT NULL, protracted INT NOT NULL, "
		sql_query += "min_speciation_gen DOUBLE NOT NULL, max_speciation_gen DOUBLE NOT NULL, dispersal_map TEXT NOT NULL);"
		c.execute(sql_query)
		try:
			sql_query = "INSERT INTO SIMULATION_PARAMETERS(seed, job_type, output_dir, speciation_rate, sigma, tau, deme," \
						" sample_size, max_time, dispersal_relative_cost, min_num_species, habitat_change_rate, gen_since_historical," \
						" time_config_file, coarse_map_file, coarse_map_x, coarse_map_y, coarse_map_x_offset, " \
						"coarse_map_y_offset, coarse_map_scale, fine_map_file, fine_map_x, fine_map_y, fine_map_x_offset," \
						"fine_map_y_offset, sample_file, grid_x, grid_y, sample_x, sample_y, sample_x_offset, sample_y_offset, " \
						"historical_coarse_map, historical_fine_map, sim_complete, dispersal_method, m_probability, cutoff, " \
						"restrict_self, landscape_type, protracted, min_speciation_gen, max_speciation_gen, dispersal_map)" \
						" SELECT seed, job_type, output_dir, speciation_rate, sigma, tau, deme," \
						" sample_size, max_time, dispersal_relative_cost, min_num_species, forest_change_rate, time_since_pristine," \
						" time_config_file, coarse_map_file, coarse_map_x, coarse_map_y, coarse_map_x_offset, " \
						"coarse_map_y_offset, coarse_map_scale, fine_map_file, fine_map_x, fine_map_y, fine_map_x_offset," \
						"fine_map_y_offset, sample_file, grid_x, grid_y, sample_x, sample_y, sample_x_offset, sample_y_offset, " \
						"pristine_coarse_map, pristine_fine_map, sim_complete, dispersal_method, m_probability, cutoff, " \
						"restrict_self, infinite_landscape, protracted, min_speciation_gen, max_speciation_gen, dispersal_map FROM SIM_P_backup;"
			c.execute(sql_query)
		except sqlite3.Error:
			# Provide additional support for a different naming convention
			try:
				sql_query = "INSERT INTO SIMULATION_PARAMETERS(seed, job_type, output_dir, speciation_rate, sigma, tau, deme," \
							" sample_size, max_time, dispersal_relative_cost, min_num_species, habitat_change_rate, gen_since_historical," \
							" time_config_file, coarse_map_file, coarse_map_x, coarse_map_y, coarse_map_x_offset, " \
							"coarse_map_y_offset, coarse_map_scale, fine_map_file, fine_map_x, fine_map_y, fine_map_x_offset," \
							"fine_map_y_offset, sample_file, grid_x, grid_y, sample_x, sample_y, sample_x_offset, sample_y_offset, " \
							"historical_coarse_map, historical_fine_map, sim_complete, dispersal_method, m_probability, cutoff, " \
							"restrict_self, landscape_type, protracted, min_speciation_gen, max_speciation_gen, dispersal_map)" \
							" SELECT seed, job_type, output_dir, speciation_rate, sigma, tau, deme," \
							" sample_size, max_time, dispersal_relative_cost, min_num_species, habitat_change_rate, time_since_pristine," \
							" time_config_file, coarse_map_file, coarse_map_x, coarse_map_y, coarse_map_x_offset, " \
							"coarse_map_y_offset, coarse_map_scale, fine_map_file, fine_map_x, fine_map_y, fine_map_x_offset," \
							"fine_map_y_offset, sample_file, grid_x, grid_y, sample_x, sample_y, sample_x_offset, sample_y_offset, " \
							"pristine_coarse_map, pristine_fine_map, sim_complete, dispersal_method, m_probability, cutoff, " \
							"restrict_self, infinite_landscape, protracted, min_speciation_gen, max_speciation_gen, dispersal_map FROM SIM_P_backup;"
				c.execute(sql_query)
			except sqlite3.Error:
				# Provide additional support for a different naming convention
				sql_query = "INSERT INTO SIMULATION_PARAMETERS(seed, job_type, output_dir, speciation_rate, sigma, tau, deme," \
							" sample_size, max_time, dispersal_relative_cost, min_num_species, habitat_change_rate, gen_since_historical," \
							" time_config_file, coarse_map_file, coarse_map_x, coarse_map_y, coarse_map_x_offset, " \
							"coarse_map_y_offset, coarse_map_scale, fine_map_file, fine_map_x, fine_map_y, fine_map_x_offset," \
							"fine_map_y_offset, sample_file, grid_x, grid_y, sample_x, sample_y, sample_x_offset, sample_y_offset, " \
							"historical_coarse_map, historical_fine_map, sim_complete, dispersal_method, m_probability, cutoff, " \
							"restrict_self, landscape_type, protracted, min_speciation_gen, max_speciation_gen, dispersal_map)" \
							" SELECT seed, job_type, output_dir, speciation_rate, sigma, tau, deme," \
							" sample_size, max_time, dispersal_relative_cost, min_num_species, habitat_change_rate, gen_since_pristine," \
							" time_config_file, coarse_map_file, coarse_map_x, coarse_map_y, coarse_map_x_offset, " \
							"coarse_map_y_offset, coarse_map_scale, fine_map_file, fine_map_x, fine_map_y, fine_map_x_offset," \
							"fine_map_y_offset, sample_file, grid_x, grid_y, sample_x, sample_y, sample_x_offset, sample_y_offset, " \
							"pristine_coarse_map, pristine_fine_map, sim_complete, dispersal_method, m_probability, cutoff, " \
							"restrict_self, infinite_landscape, protracted, min_speciation_gen, max_speciation_gen, dispersal_map FROM SIM_P_backup;"
				c.execute(sql_query)
		c.execute("DROP TABLE SIM_P_backup;")
		db.commit()
		db.close()
		db = None
	except sqlite3.Error as soe:  # pragma: no cover
		raise sqlite3.Error("Could not update simulation parameters: {}".format(soe))



"""
Safely open, close and fetch data from an sqlite connection.

:class:`.SQLiteConnection` contains context management for opening sql connections, plus basic functionality for
detecting existence and structure of databases.
"""
import sqlite3


class SQLiteConnection():
	"""
	Class containing context management for opening sqlite3 connections. The file name provided can either be a string
	containing the path to the file, or an sqlite3.Connection object, which will NOT be closed on destruction. This
	provides two points of entry to the system with the same interface.
	"""
	def __init__(self, filename):
		"""
		Initialises the SQLiteConnection object.

		:param filename: Should be either a string containing a path to a file to open, or an already-open
						 sqlite3.Connection object. If the latter, the connection will not be closed on destruction.
					 	 filename can also be a sqlite3.Cursor object, in which case it will simply be returned.
		"""
		self.filename = filename
		self.opened_here = False
		self.database = None

	def __enter__(self):
		"""
		Opens the connection to the file and returns an active cursor.
		:return:
		"""
		if isinstance(self.filename, sqlite3.Connection):
			self.database = self.filename
		elif isinstance(self.filename, sqlite3.Cursor):
			return self.filename
		else:
			self.opened_here = True
			try:
				self.database = sqlite3.connect(self.filename)
			except sqlite3.Error as soe:  # pragma: no cover
				self.opened_here = False
				raise soe
		return self.database.cursor()

	def __exit__(self, *args):
		if self.opened_here:
			self.database.commit()
			self.database.close()
			self.database = None



def check_sql_table_exist(database, table_name):
	"""
	Checks that the supplied table exists in the supplied database.

	:param database: the database to check existence in
	:param table_name: the table name to check for

	:return: true if the table exists
	:rtype: bool
	"""
	with SQLiteConnection(database) as cursor:
		table_names = [x[0] for x in cursor.execute("SELECT name FROM sqlite_master WHERE type='table'").fetchall()]
	return table_name in table_names

def check_sql_column_exists(database, table_name, column_name):
	"""
	Checks if the column exists in the database.

	:param database: the database to check existence in
	:param table_name: the table name to check within
	:param column_name: the column name to check for

	:return: true if the column exists.
	:rtype: bool
	"""
	with SQLiteConnection(database) as cursor:
		c = [i[1] for i in cursor.execute('PRAGMA table_info({})'.format(table_name)) if i[1] == column_name]
		return len(c) != 0

def fetch_table_from_sql(database, table_name):
	"""
	Returns a list of the data contained by the provided table in the database.

	:raises sqlite3.Error: if the table is not contained in the database (protects SQL injections).

	:param database: the database to obtain from
	:param table_name: the table name to fetch data from
	:return: a list of lists, containing all data within the provided table in the database
	"""
	try:
		with SQLiteConnection(database) as c:
			if not check_sql_table_exist(c, table_name):
				raise sqlite3.Error("Table {} does not exist in database.".format(table_name))
			c.execute("SELECT * FROM {}".format(table_name))
			return [list(x) for x in c.fetchall()]
	except sqlite3.Error as soe:  # pragma: no cover
		raise IOError("Cannot fetch {} from database {}: {}".format(table_name, database, soe))



def sql_get_max_from_column(database, table_name, column_name):
	"""
	Returns the maximum value from the specified column.

	:param database: the database to fetch from
	:param table_name: the table name to attain
	:param column_name:
	:return:
	"""
	with SQLiteConnection(database) as cursor:
		if not check_sql_table_exist(cursor, table_name):
			raise sqlite3.Error("Table {} does not exist in database.".format(table_name))
		if not check_sql_column_exists(cursor, table_name, column_name):
			raise sqlite3.Error("Column {} does not exist in {}.".format(column_name, table_name))
		ex_str = "SELECT max({}) FROM {}".format(column_name, table_name)
		cursor.execute(ex_str)
		return cursor.fetchone()[0]



"""
Compile **necsim** with default or provided compilation options. Intended for internal usage during ``pip`` or ``conda``
builds, although manual installation is also possible by running this file from the command line.
``python installer.py`` configures the install by detecting system components and compiles the ``C++`` files, if
possible. Command line flags can be provided to installer.py to modify the install (see
:ref:`Compilation Options <sec Compilation Options>` for more information).
"""

from __future__ import print_function, absolute_import, division  # Only Python 2.x

import itertools
import logging
import os
import platform
import re
import shutil
import subprocess
import sys
import time
from distutils import sysconfig

# Import system operations for subprocess executation and log file handling

try:
	from .future_except import FileNotFoundError
	from .system_operations import execute_log_info, set_logging_method, execute_silent, mod_directory
except (ImportError, SystemError, ValueError):  # pragma: no cover
	from future_except import FileNotFoundError
	from system_operations import execute_log_info, set_logging_method, execute_silent, mod_directory
from setuptools.command.build_ext import build_ext


class Installer(build_ext):  # pragma: no cover
	"""Wraps configuration and compilation of C++ code."""

	def __init__(self, dist, **kwargs):
		"""Generates the link to the mod directory for installation."""
		build_ext.__init__(self, dist)
		self.mod_dir = mod_directory
		self.build_dir = None
		self.obj_dir = None

	def make_depend(self):
		"""
		Runs make depend in the lib directory to calculate all dependencies for the header and source files.

		.. note:: Fails silently if makedepend is not installed, printing an error to logging.
		"""
		try:
			execute_silent(["make", "depend"], cwd=os.path.join(self.mod_dir, "lib/"))
		except (RuntimeError, subprocess.CalledProcessError) as rte:
			logging.error("Could not execute makedepend: " + str(rte))
			logging.error("Using default dependencies instead.")
			self.use_default_depends()

	def create_default_depend(self):
		"""
		Runs the default makedepend command, outputting dependencies to lib/depends_default.

		Used to generate a default dependency file on a system where makedepend exists, for a system where it does not.
		"""
		try:
			execute_silent(["makedepend", "*.cpp", "necsim/*.cpp", "-f" "depends_default", "-p", "obj/", "-Y"],
						   cwd=os.path.join(self.mod_dir, "lib/"))
		except (RuntimeError, subprocess.CalledProcessError) as rte:
			logging.error("Could not execute makedepend: " + str(rte))

	def use_default_depends(self):
		"""
		Uses the default dependencies, copying all contents of depends_default to the end of Makefile.

		.. note:: Zero error-checking is done here as the Makefiles should not change, and the depends_default file should
				  be created using create_default_depend()

		"""
		# First read Makefile
		with open(os.path.join(self.mod_dir, "lib", "Makefile"), "r") as f_in:
			lines = f_in.readlines()
		with open(os.path.join(self.mod_dir, "lib", "depends_default"), "r") as f_default:
			lines_default = f_default.readlines()
		# Now write all back out
		with open(os.path.join(self.mod_dir, "lib/Makefile"), "w") as f_out:
			for line in lines:
				if "# DO NOT DELETE" in line:
					break
				f_out.write(line)
			for line in lines_default:
				f_out.write(line)

	def do_compile(self):
		"""
		Compiles the C++ necsim program by running make. This changes the working directory to wherever the module has been
		installed for the subprocess call.
		"""
		# Check that the make file exists
		if os.path.exists(os.path.join(self.mod_dir, "lib/")):
			time.sleep(0.5)
			self.make_depend()
			# Sleep to ensure that file timings are updated (support for HPC systems with inaccurate timings).
			time.sleep(1)
			try:
				execute_log_info(["make", "all"], cwd=os.path.join(self.mod_dir, "lib/"))
				logging.info("Compilation exited successfully.")
			except (RuntimeError, subprocess.CalledProcessError) as rte:
				logging.error(str(rte))
				logging.error("Compilation attempted, but error thrown.")
		else:
			raise IOError("C++ library does not exist! Check relative file path")

	def move_shared_object_file(self):
		"""
		Moves the shared object (.so) file to the build directory.
		:return:
		"""
		directory = os.path.join(self.mod_dir, "necsim")
		dirv = 'sharedpy' + sys.version[0]
		for file in ["libnecsim.so"]:
			src = os.path.join(self.mod_dir, "lib", file)
			version_dir = os.path.join(directory, dirv)
			if not os.path.exists(src):
				raise IOError("Shared object file {} does not exists. Check installation was successful.".format(src))
			if not os.path.exists(directory):
				os.mkdir(directory)
			if not os.path.exists(version_dir):
				os.mkdir(version_dir)
			if not os.path.exists(os.path.join(version_dir, '__init__.py')):
				open(os.path.join(version_dir, '__init__.py'), 'a').close()
			dst = os.path.join(directory, dirv, file)
			if os.path.exists(dst):
				os.remove(dst)
			shutil.copy(src, dst)

	def get_build_dir(self):
		"""
		Gets the build directory.

		:return: the build directory path
		"""
		if self.build_dir is None:
			directory = os.path.join(self.mod_dir, "necsim")
			return directory
		return os.path.join(self.build_dir)

	def get_obj_dir(self):
		"""
		Gets the obj directory for installing obj files to.

		:return: the obj directory path
		"""
		if self.obj_dir is None:
			return "obj"
		return self.obj_dir

	def configure(self, opts=None):
		"""
		Runs ./configure --opts with the supplied options. This should create the makefile for compilation, otherwise a
		RuntimeError will be thrown.

		:param opts: a list of options to pass to the ./configure call
		"""
		if "--with-debug" in opts:
			for i, each in enumerate(opts):
				if "-DNDEBUG" in each:
					opts[i] = opts[i].replace("-DNDEBUG", "")
		if os.path.exists(os.path.join(self.mod_dir, "lib", "configure")):
			try:
				if opts is None:
					execute_log_info(["./configure", "--with-verbose", "BUILDDIR={}".format(self.get_build_dir()),
									  "OBJDIR={}".format(self.get_obj_dir())],
									 cwd=os.path.join(self.mod_dir, "lib"))
				else:
					command = ["./configure"]
					command.extend(opts)
					if "BUILDDIR" not in opts:
						command.append("BUILDDIR={}".format(self.get_build_dir()))
					if "OBJDIR" not in opts:
						command.append("OBJDIR={}".format(self.get_obj_dir()))
					execute_log_info(command, cwd=os.path.join(self.mod_dir, "lib"))
			except subprocess.CalledProcessError as cpe:
				raise subprocess.CalledProcessError("Configuration attempted, but error thrown: ".format(cpe))
		else:
			raise RuntimeError("File src/configure does not exist. Check installation has been successful.")

	def autoconf(self):
		"""
		Runs the `autoconf` bash function (assuming that autoconf is available) to create the `configure` executable.
		"""
		try:
			execute_log_info(["autoconf"], cwd=os.path.join(self.mod_dir, "lib/"))
		except (RuntimeError, subprocess.CalledProcessError, FileNotFoundError, OSError, IOError) as cpe:
			logging.warning("Could not run autoconf function to generate configure executable. "
							"Please run this functionality manually if installation fails.")
			logging.warning(str(cpe))

	def clean(self):
		"""
		Runs make clean in the NECSim directory to wipe any previous potential compile attempts.
		"""
		try:
			time.sleep(0.5)
			execute_log_info(["make", "obj_dir"], cwd=os.path.join(self.mod_dir, "lib/"))
			execute_log_info(["make", "build_dir"], cwd=os.path.join(self.mod_dir, "lib/"))
			execute_log_info(["make", "clean"], cwd=os.path.join(self.mod_dir, "lib/"))
		except subprocess.CalledProcessError as cpe:
			raise RuntimeError("Make file has not been generated. Cannot clean: ".format(cpe))

	def get_ldflags(self):
		"""Get the ldflags that Python was compiled with, removing some problematic options."""
		sysldflags = sysconfig.get_config_var("LDFLAGS")
		if sysldflags is None:
			ldflags = ""
		else:
			ldflags = re.sub(r"-arch \b[^ ]*[\ ]*", "", sysldflags) + " "
		if "--sysroot=" in ldflags:
			logging.warning("--sysroot found in LDFLAGS, removing")
			ldflags = re.sub(r"--sysroot=.*[,\s]", "", ldflags)
		ldflags = ldflags.replace("\n", " ")
		return ldflags

	def get_ldshared(self):
		"""Get the ldshared Python variables and replaces -bundle with -shared for proper compilation."""
		ldflags = sysconfig.get_config_var("LDSHARED")
		if ldflags is None:
			return ""
		ldflags = " ".join(ldflags.split()[1:]).replace("-bundle", "-shared")
		return ldflags

	def get_compilation_flags(self, display_warnings=False):
		"""
		Generates the compilation flags for passing to ./configure.
		:param display_warnings: If true, runs with the -Wall flag for compilation (displaying all warnings). Default is False.

		:return: list of compilation flags.
		:rtype: list
		"""
		# Get the relevant flags that Python was originally compiled with, to be passed to the C++ code.
		include = str("CPPFLAGS=-I" + sysconfig.get_python_inc()).replace("\n", "")
		cflags = " " + sysconfig.get_config_var('CFLAGS')
		cflags = str(re.sub(r"-arch \b[^ ]*", "", cflags)).replace("\n", "")  # remove any architecture flags
		cflags += " "
		py_ldflags = str("-L" + sysconfig.get_python_lib(standard_lib=True) +
						 " -L" + sysconfig.get_config_var('DESTDIRS').replace(" ", " -L")).replace("\n", "")
		py_lib = "PYTHON_LIB=-lpython"
		ldflags = str("LDFLAGS=" + self.get_ldflags())
		# Get the shared object platform-specific compilation flags.
		platform_so = "PLATFORM_SO="
		if platform.system() == "Linux":
			platform_so += "-shared"
		elif platform.system() == "Darwin":
			platform_so += "-dynamiclib"
		elif platform.system() == "Windows":
			raise SystemError(
				"COMPILATION FAILURE: Windows is not yet supported. You must compile the libraries yourself.")
		else:
			logging.critical("OS not detected, compilation failures likely. Please report this error.")
		# Make sure that the linker directs to the correct Python library (such as -lpython3.5m)
		# Eventually this will also detect if the install is for an anaconda distribution.
		if 'conda' not in sys.version and 'Continuum' not in sys.version:
			py_lib += sys.version[0:3]
			if 'm' in sysconfig.get_config_var('LIBRARY'):
				py_lib += 'm'
		else:
			cflags += "-I{}/include {}".format(os.environ["PREFIX"], "--enable-shared")
			ldflags += "-L{}/lib".format(os.environ["PREFIX"])
			py_lib += sys.version[0:3]
			if 'm' in sysconfig.get_config_var('LIBRARY'):
				py_lib += 'm'
			cflags += " -DANACONDA"
		ldflags += " " + py_ldflags
		py_ldflags = "PYTHON_LDFLAGS=" + py_ldflags
		call = [include + cflags, py_lib, ldflags, py_ldflags, platform_so]
		# Remove the flags which would potentially cause unnecessary warnings to be thrown.
		# This can be disabled by passing display_warnings=True
		if not display_warnings:
			call = [re.sub('-Wstrict-prototypes|-Wno-unused-result|-Wunused-variable|-Wall', '', x) for x in call]
		return call

	def run_configure(self, argv=None, logging_level=logging.INFO, display_warnings=False):
		"""
		Configures the install for compile options provided via the command line, or with default options if no options exist.
		Running with ``-help`` or `-h` will display the compilation configurations called from ``./configure``.

		:param argv: the arguments to pass to configure script
		:param logging_level: the logging level to utilise (defaults to INFO).
		:param display_warnings: If true, runs with the -Wall flag for compilation (displaying all warnings). Default is False.
		"""
		if argv is None:
			argv = [None]
		call = self.get_compilation_flags(display_warnings=display_warnings)
		set_logging_method(logging_level=logging_level)
		self.autoconf()
		if len(argv) == 1:
			logging.info("No compile options provided, using defaults.")
			print("Obj: {}".format(self.get_obj_dir()))
			print("Build: {}".format(self.get_build_dir()))
			call.extend(["--with-verbose", "OBJDIR={}", "BUILDDIR={}".format(self.get_obj_dir(),
																			 self.get_build_dir())])
		else:
			if argv[1] == "--h" or argv[1] == "-h" or argv[1] == "-help" or argv[1] == "--help":
				execute_log_info(["./configure", "--help"], cwd=os.path.join(self.mod_dir, "lib/"))
			if isinstance(argv, str):
				call.append(argv[2])
			else:
				call.extend(argv[1:])
		logging.info(call)
		self.configure(opts=call)
		self.clean()

	def backup_makefile(self):
		"""
		Copies the makefile to a saved folder so that even if the original is overwritten, the last successful
		compilation can be recorded.
		"""
		src = os.path.join(self.mod_dir, "lib", "Makefile")
		dst = os.path.join(self.mod_dir, "reference", "Makefile")
		if not os.path.exists(src):
			logging.error("Makefile does not exist at {}. Check successful compilation".format(src))
			return
		if os.path.exists(dst):
			os.remove(dst)
		shutil.copy2(src, dst)

	def copy_makefile(self):
		"""
		Copies the backup makefile to the main directory, if it exists.
		Throws an IOError if no makefile is found.
		"""
		src = os.path.join("reference", "Makefile")
		dst = os.path.join("lib", "Makefile")
		if not os.path.exists(src):
			raise IOError("Cannot find backup Makefile, requires running of configuration scripts.")
		if os.path.exists(dst):
			os.remove(dst)
		shutil.copy2(src, dst)

	def configure_and_compile(self, argv=[None], logging_level=logging.INFO):
		"""
		Calls the configure script, then runs the compilation.

		:param argv: the arguments to pass to configure script
		:param logging_level: the logging level to utilise (defaults to INFO).
		:rtype: None
		"""
		set_logging_method(logging_level=logging_level)
		display_warnings = "display_warnings=True" in argv
		self.run_configure(argv, logging_level=logging_level, display_warnings=display_warnings)
		self.do_compile()
		# move_shared_object_file()
		# move_executable()
		self.backup_makefile()

	def setuptools_cmake(self, ext):
		"""
		Configures cmake for setuptools usage.

		:param ext: the extension to build cmake on
		"""
		if 'conda' not in sys.version and 'Continuum' not in sys.version:
			extdir = os.path.abspath(os.path.join(os.path.dirname(self.get_ext_fullpath(ext.name)), "pycoalescence",
												  "necsim"))
		else:
			sp_dir = os.environ.get("SP_DIR")
			if sp_dir is None:
				sp_dir = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name)))
			extdir = os.path.join(sp_dir, "pycoalescence", "necsim")
		cmake_args, build_args = self.get_default_cmake_args(extdir)
		env = os.environ.copy()
		env['CXXFLAGS'] = '{} -DVERSION_INFO=\\"{}\\"'.format(
			env.get('CXXFLAGS', ''),
			self.distribution.get_version())
		if "INTEL_LICENSE_FILE" in env.keys():
			env["CXX"] = "icpc"
			env["CC"] = "icc"
			cmake_args.extend(["-DCMAKE_C_COMPILER=icc", "-DCMAKE_CXX_COMPILER=icpc", "-DUSING_INTEL=ON"])
		self.run_cmake(ext.sourcedir, cmake_args, build_args, self.build_temp, env)

	def run_cmake(self, src_dir, cmake_args, build_args, tmp_dir, env):
		"""
		Runs cmake to compile necsim.

		:param src_dir: the source directory for necsim .cpp and .h files
		:param cmake_args: arguments to pass to the cmake project
		:param tmp_dir: the build directory to output cmake files to
		:param env: the os.environ (or other environmental variables) to pass on
		"""
		if not os.path.exists(tmp_dir):
			os.makedirs(tmp_dir)
		try:
			subprocess.check_call(['cmake', src_dir] + cmake_args,
								  cwd=tmp_dir, env=env)
			subprocess.check_call(['cmake', '--build', ".", "--target", "necsim"] + build_args,
								  cwd=tmp_dir)
		except subprocess.CalledProcessError as cpe:
			raise SystemError("Fatal error running cmake in directory: {}".format(cpe))
		if platform.system() == "Windows":
			shutil.copy(os.path.join(tmp_dir, "Release", "necsim.pyd"), os.path.join(self.get_build_dir(),
																					 "libnecsim.pyd"))

	def run(self):
		"""Runs installation and generates the shared object files - entry point for setuptools"""
		for ext in self.extensions:
			self.build_extension(ext)

	def build_extension(self, ext):
		"""Builds the C++ and Python extension."""
		self.build_dir = os.path.abspath(os.path.join(os.path.dirname(self.get_ext_fullpath(ext.name)),
													  "pycoalescence", "necsim"))
		self.obj_dir = self.build_dir
		# if platform.system() == "Windows":
		# 	raise SyntaxError("Windows is not supported by pycoalescence at this time.")
		if not os.path.exists(self.build_temp):
			os.makedirs(self.build_temp)
		self.setuptools_cmake(ext)

	def get_default_cmake_args(self, output_dir):
		"""
		Returns the default cmake configure and build arguments.

		:param output_dir: the output directory to use

		:return: tuple of two lists, first containing cmake configure arguments, second containing build arguments
		:rtype: tuple
		"""
		cfg = 'Debug' if self.debug else 'Release'
		# pymalloc_ext = "m" if bool(sysconfig.get_config_var('WITH_PYMALLOC')) else ""
		# v_info = sys.version_info
		cflags = sysconfig.get_config_var('CFLAGS')
		if cflags is not None:
			cflags = str(re.sub(r"-arch \b[^ ]*", "", cflags)).replace("\n", "")  # remove any architecture flags
		else:
			cflags = ""
		conda_prefix = os.environ.get("PREFIX", None) # for conda only
		gdal_inc_path = None
		gdal_dir = None
		if platform.system() == "Windows":
			libdir = get_python_library("{}.{}".format(sys.version_info.major, sys.version_info.minor))
			if conda_prefix is not None:
				gdal_inc_path = os.path.join(conda_prefix, "Library", "include") # TODO check this works - this is the bug - change to gdal working location on Windows
				gdal_dir = os.path.join(conda_prefix, "Library", "lib")
		else:
			libdir = sysconfig.get_config_var("LIBDIR")
			if conda_prefix is not None:
				gdal_inc_path = os.path.join(conda_prefix, "include")
				gdal_dir = os.path.join(conda_prefix, "lib")
			else:
				try:
					gdal_inc_path = subprocess.check_output(["gdal-config","--cflags"])
					gdal_dir = subprocess.check_output(["gdal-config", "--prefix"])
				except subprocess.CalledProcessError:
					pass
				if gdal_inc_path is not None:
					gdal_inc_path = gdal_inc_path.decode("utf-8").split(" ")[0][2:].replace("\n", "")
				if gdal_dir is not None:
					gdal_dir = gdal_dir.decode("utf-8").replace("\n", "")
		if libdir is None:
			libdir = os.path.abspath(os.path.join(sysconfig.get_config_var('LIBDEST'), "..", "libs"))
			if sysconfig.get_config_var("LIBDEST") is None:
				raise SystemError("Cannot detect library directory for Python install.")
		cmake_args = ["-DPYTHON_LIBRARY:FILEPATH={}".format(libdir),
					  # "-DPYTHON_LINKER:=-lpython{}.{}{}".format(v_info.major, v_info.minor, pymalloc_ext),
					  "-DPYTHON_CPPFLAGS:='{}'".format(cflags),
					  "-DPYTHON_LDFLAGS:='{}'".format(self.get_ldshared()),
					  # "-DPYTHON_LIBRARY:FILEPATH={}".format(get_python_library("{}.{}".format(sys.version_info.major,
					  # 																				  sys.version_info.minor))),
					  "-DPYTHON_INCLUDE_DIR:FILEPATH={}".format(sysconfig.get_python_inc()),
					  '-DCMAKE_BUILD_TYPE={}'.format(cfg)]
		if gdal_inc_path is not None:
			cmake_args.append("-DGDAL_INCLUDE_DIR={}".format(gdal_inc_path))
		if gdal_dir is not None:
			cmake_args.append("-DGDAL_DIR={}".format(gdal_dir))
		build_args = ['--config', cfg]
		if platform.system() == "Windows":
			cmake_args += ['-DCMAKE_LIBRARY_OUTPUT_DIRECTORY:PATH={}'.format(
				# cfg.upper(),
				output_dir)]
			if sys.maxsize > 2 ** 32:
				cmake_args += ['-A', 'x64']
			build_args += ['--', '/m']
		else:

			cmake_args += ['-DCMAKE_BUILD_TYPE=' + cfg,
						   '-DCMAKE_LIBRARY_OUTPUT_DIRECTORY:PATH={}'.format(output_dir)]
			build_args += ['--', '-j2']
		return cmake_args, build_args


def get_python_library(python_version):  # pragma: no cover
	"""Get path to the Python library associated with the current Python
	interpreter."""
	# determine direct path to libpython
	python_library = sysconfig.get_config_var('LIBRARY')
	potential_library = None
	# if static (or nonexistent), try to find a suitable dynamic libpython
	if python_library is None or python_library[-2:] == '.a':

		candidate_lib_prefixes = ['', 'lib']

		candidate_extensions = ['.lib', '.so']
		if sysconfig.get_config_var('WITH_DYLD'):
			candidate_extensions.insert(0, '.dylib')

		candidate_versions = [python_version]
		if python_version:
			candidate_versions.append('')
			candidate_versions.insert(
				0, "".join(python_version.split(".")[:2]))

		abiflags = getattr(sys, 'abiflags', '')
		candidate_abiflags = [abiflags]
		if abiflags:
			candidate_abiflags.append('')

		# Ensure the value injected by virtualenv is
		# returned on windows.
		# Because calling `sysconfig.get_config_var('multiarchsubdir')`
		# returns an empty string on Linux, `du_sysconfig` is only used to
		# get the value of `LIBDIR`.
		libdir = sysconfig.get_config_var('LIBDIR')
		if sysconfig.get_config_var('MULTIARCH'):
			masd = sysconfig.get_config_var('multiarchsubdir')
			if masd:
				if masd.startswith(os.sep):
					masd = masd[len(os.sep):]
				libdir = os.path.join(libdir, masd)

		if libdir is None:
			libdir = os.path.abspath(os.path.join(
				sysconfig.get_config_var('LIBDEST'), "..", "libs"))

		candidates = (
			os.path.join(
				libdir,
				''.join((pre, 'python', ver, abi, ext))
			)
			for (pre, ext, ver, abi) in itertools.product(
			candidate_lib_prefixes,
			candidate_extensions,
			candidate_versions,
			candidate_abiflags
		)
		)
		for candidate in candidates:
			if os.path.exists(candidate):
				# we found a (likely alternate) libpython
				potential_library = candidate
				if potential_library[-2:] != ".a":
					break
	# Otherwise still a static library, keep searching
	if potential_library is None:
		raise IOError("No Python libraries found")
	return potential_library


if __name__ == "__main__":  # pragma: no cover
	fail = True
	from distutils.dist import Distribution
	import argparse

	parser = argparse.ArgumentParser(description='Build the C++ library (necsim) required for pycoalescence.')
	parser.add_argument('--cmake', action='store_true', default=True, dest="cmake",
						help='use the cmake build process')
	parser.add_argument('--autotools', action='store_true', default=False, dest="autotools",
						help='Use the autotools build process (./configure and make)')
	parser.add_argument('--compiler-args', metavar='N', type=str, nargs='+', dest="compiler_args",
						default=[],
						help='Additional arguments to pass to the autotools compiler')
	parser.add_argument('--cmake-args', metavar='N', type=str, nargs='+', dest="cmake_args",
						default=[],
						help='Additional arguments to pass to the cmake compiler during configuration')
	parser.add_argument('--cmake-build-args', metavar='N', type=str, nargs='+', dest="cmake_build_args",
						default=[],
						help='Additional arguments to pass to the cmake compiler at build time')
	parser.add_argument('--debug', action='store_false', default=False, dest='debug',
						help='Compile using DEBUG defines')
	parser.add_argument('-c', '-C', '--compile', action='store_true', default=False, dest='compile_only',
						help='Compile only, do not re-configure necsim')

	args = parser.parse_args()
	if args.cmake and args.autotools:
		raise ValueError("Cannot use both cmake and autotools build process - specify one or the other.")
	if not args.cmake or args.autotools:
		raise ValueError("Must specify compilation either using autotools or cmake.")
	dist = Distribution()
	installer = Installer(dist)
	if args.cmake:
		env = os.environ.copy()
		build_dir = installer.get_build_dir()
		obj_dir = installer.get_obj_dir()
		cmake_args, build_args = installer.get_default_cmake_args(build_dir)
		cmake_args += args.cmake_args
		build_args += args.cmake_build_args
		src_dir = os.path.join(installer.mod_dir, "lib")
		installer.run_cmake(src_dir, cmake_args, build_args, obj_dir, env)
	else:
		if platform.system() != "Windows":
			raise SystemError("Usage of configure and make on a windows system is not supported.")
		if args.compile_only:
			set_logging_method(logging_level=logging.INFO)
			installer.copy_makefile()
			installer.do_compile()
			fail = False
		if fail:
			installer.configure_and_compile(args.compiler_args)



"""
Contains compatibility imports for Python 2.x.
"""

try:  # pragma: no cover
	FileExistsError = FileExistsError
	FileNotFoundError = FileNotFoundError
except NameError:  # pragma: no cover
	class FileExistsError(IOError):
		pass

	class FileNotFoundError(IOError):
		pass